/* 
 * Copyright (C) 2004      Disch
 * Copyrigth (C) 2006      hcs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  NSF_File.cpp
//
//


#include <stdio.h>
#include <stdlib.h>
#include "nsf_types.h"
#include "NSF_Core.h"
#include "NSF_File.h"

typedef int bool;

///////////////////////////////////
// (global) data members

//basic NSF info
bool				bIsExtended=0;		//0 = NSF, 1 = NSFE
BYTE				nIsPal=0;				//0 = NTSC, 1 = PAL, 2,3 = mixed NTSC/PAL (interpretted as NTSC)
int					nfileLoadAddress=0;		//The address to which the NSF code is loaded to
int					nfileInitAddress=0;		//The address of the Init routine (called at track change)
int					nfilePlayAddress=0;		//The address of the Play routine (called several times a second)
BYTE				nChipExtensions=0;	//Bitwise representation of the external chips used by this NSF.  Read NSFSpec.txt for details.
	
//old NESM speed stuff (blarg)
int					nNTSC_PlaySpeed=0;
int					nPAL_PlaySpeed=0;

//track info
int					nTrackCount=0;		//The number of tracks in the NSF (1 = 1 track, 5 = 5 tracks, etc)
int					nInitialTrack=0;		//The initial track (ZERO BASED:  0 = 1st track, 4 = 5th track, etc)

//nsf data
BYTE*				pDataBuffer=0;		//the buffer containing NSF code.  If needdata was false when loading the NSF, this is NULL
int					nDataBufferSize=0;	//the size of the above buffer.  0 if needdata was false

//playlist
BYTE*				pPlaylist=0;			//the buffer containing the playlist (NULL if none exists).  Each entry is the zero based index of the song to play
int					nPlaylistSize=0;		//the size of the above buffer (and the number of tracks in the playlist)

//track time / fade
int*				pTrackTime=0;			//the buffer containing the track times.  NULL if no track times specified.  Otherwise this buffer MUST BE (nTrackCount * sizeof(int)) in size
int*				pTrackFade=0;			//the buffer containing the track fade times.  NULL if none are specified.  Same conditions as pTrackTime

//track labels
char**				szTrackLabels=0;		//the buffer containing track labels.  NULL if there are no labels.  There must be nTrackCount char pointers (or none if NULL).
										//Each pointer must point to it's own buffer containing a string (the length of this buffer doesn't matter, just so long as the string is NULL terminated)
										// the string's buffer may be NULL if a string isn't needed
										//szTrackLabels as well as all of the buffers it points to are destroyed upon
										// a call to Destroy (or the destructor).

//string info
char*				szGameTitle=0;		//pointer to a NULL-terminated string containing the name of the game.  Or NULL if no name specified
char*				szArtist=0;			//pointer to a NULL-terminated string containing the author of the NSF.  Or NULL if none specified
char*				szCopyright=0;		//pointer to a NULL-terminated string containing the copyright info.  Or NULL if none specified
char*				szRipper=0;			//pointer to a NULL-terminated string containing the 'hacker' who ripped the NSF.  Or NULL if none specified

//bankswitching info
BYTE				nBankswitch[8]={0};		//The initial bankswitching registers needed for some NSFs.  If the NSF does not use bankswitching, these values will all be zero


#define SAFE_DELETE(p) { if(p){ free( p ); p = NULL; } }
#define SAFE_NEW(p,t,s,r) p = malloc(sizeof(t)*(s)); if(!p) return r; ZeroMemory(p,sizeof(t)*(s))


int		LoadFile(LPCSTR path,BYTE needdata,BYTE ignoreversion)
{
	NSFFile_Destroy();

	FILE* file = fopen(path,"rb");
	if(!file) return -1;

	UINT type = 0;
	fread(&type,4,1,file);
	int ret = -1;

	if(type == HEADERTYPE_NESM)		ret = LoadFile_NESM(file,needdata,ignoreversion);
	if(type == HEADERTYPE_NSFE)		ret = LoadFile_NSFE(file,needdata);

	fclose(file);

	// Snake's revenge puts '00' for the initial track, which (after subtracting 1) makes it 256 or -1 (bad!)
	// This prevents that crap
	if(nInitialTrack >= nTrackCount)
		nInitialTrack = 0;
	if(nInitialTrack < 0)
		nInitialTrack = 0;

	// if there's no tracks... this is a crap NSF
	if(nTrackCount < 1)
	{
		NSFFile_Destroy();
		return -1;
	}

	return ret;
}

void	NSFFile_Destroy()
{
	SAFE_DELETE(pDataBuffer);
	SAFE_DELETE(pPlaylist);
	SAFE_DELETE(pTrackTime);
	SAFE_DELETE(pTrackFade);
	if(szTrackLabels)
	{
	    int i;
		for(i = 0; i < nTrackCount; i++)
			SAFE_DELETE(szTrackLabels[i]);
		SAFE_DELETE(szTrackLabels);
	}
	SAFE_DELETE(szGameTitle);
	SAFE_DELETE(szArtist);
	SAFE_DELETE(szCopyright);
	SAFE_DELETE(szRipper);

	//ZeroMemory(this,sizeof(CNSFFile));
}

int LoadFile_NESM(FILE* file,BYTE needdata,BYTE ignoreversion)
{
	int len;

	fseek(file,0,SEEK_END);
	len = ftell(file) - 0x80;
	fseek(file,0,SEEK_SET);

	if(len < 1) return -1;

	//read the info
	struct NESM_HEADER					hdr;
	fread(&hdr,0x80,1,file);

	//confirm the header
	if(hdr.nHeader != HEADERTYPE_NESM)			return -1;
	if(hdr.nHeaderExtra != 0x1A)				return -1;
	if((!ignoreversion) && (hdr.nVersion != 1))	return -1; //stupid NSFs claim to be above version 1  >_>

	//NESM is generally easier to work with (but limited!)
	//  just move the data over from NESM_HEADER over to our member data

	bIsExtended =				0;
	nIsPal =					hdr.nNTSC_PAL & 0x03;
	nPAL_PlaySpeed =			hdr.nSpeedPAL;			//blarg
	nNTSC_PlaySpeed =			hdr.nSpeedNTSC;			//blarg
	nfileLoadAddress =				hdr.nLoadAddress;
	nfileInitAddress =				hdr.nInitAddress;
	nfilePlayAddress =				hdr.nPlayAddress;
	nChipExtensions =			hdr.nExtraChip;


	nTrackCount =				hdr.nTrackCount;
	nInitialTrack =				hdr.nInitialTrack - 1;	//stupid 1-based number =P

	memcpy(nBankswitch,hdr.nBankSwitch,8);

	SAFE_NEW(szGameTitle,char,33,1);
	SAFE_NEW(szArtist   ,char,33,1);
	SAFE_NEW(szCopyright,char,33,1);

	memcpy(szGameTitle,hdr.szGameTitle,32);
	memcpy(szArtist   ,hdr.szArtist   ,32);
	memcpy(szCopyright,hdr.szCopyright,32);

	//read the NSF data
	if(needdata)
	{
		SAFE_NEW(pDataBuffer,BYTE,len,1);
		fread(pDataBuffer,len,1,file);
		nDataBufferSize = len;
	}

	//if we got this far... it was a successful read
	return 0;
}

int LoadFile_NSFE(FILE* file,BYTE needdata)
{
	//restart the file
	fseek(file,0,SEEK_SET);

	//the vars we'll be using
	UINT nChunkType;
	int  nChunkSize;
	int  nChunkUsed;
	int  nDataPos = 0;
	BYTE	bInfoFound = 0;
	BYTE	bEndFound = 0;
	BYTE	bBankFound = 0;

	struct NSFE_INFOCHUNK	info;
	ZeroMemory(&info,sizeof(struct NSFE_INFOCHUNK));
	info.nTrackCount = 1;		//default values

	//confirm the header!
	fread(&nChunkType,4,1,file);
	if(nChunkType != HEADERTYPE_NSFE)			return -1;

	//begin reading chunks
	while(!bEndFound)
	{
		if(feof(file))							return -1;
		fread(&nChunkSize,4,1,file);
		fread(&nChunkType,4,1,file);

		switch(nChunkType)
		{
		case CHUNKTYPE_INFO:
			if(bInfoFound)						return -1;	//only one info chunk permitted
			if(nChunkSize < 8)					return -1;	//minimum size

			bInfoFound = 1;
			nChunkUsed = min((int)sizeof(struct NSFE_INFOCHUNK),nChunkSize);

			fread(&info,nChunkUsed,1,file);
			fseek(file,nChunkSize - nChunkUsed,SEEK_CUR);

			bIsExtended =			1;
			nIsPal =				info.nIsPal & 3;
			nfileLoadAddress =			info.nLoadAddress;
			nfileInitAddress =			info.nInitAddress;
			nfilePlayAddress =			info.nPlayAddress;
			nChipExtensions =		info.nExt;
			nTrackCount =			info.nTrackCount;
			nInitialTrack =			info.nStartingTrack;

			nPAL_PlaySpeed =		(WORD)(1000000 / PAL_NMIRATE);		//blarg
			nNTSC_PlaySpeed =		(WORD)(1000000 / NTSC_NMIRATE);		//blarg
			break;

		case CHUNKTYPE_DATA:
			if(!bInfoFound)						return -1;
			if(nDataPos)						return -1;
			if(nChunkSize < 1)					return -1;

			nDataBufferSize = nChunkSize;
			nDataPos = ftell(file);

			fseek(file,nChunkSize,SEEK_CUR);
			break;

		case CHUNKTYPE_NEND:
			bEndFound = 1;
			break;

		case CHUNKTYPE_TIME:
			if(!bInfoFound)						return -1;
			if(pTrackTime)						return -1;

			SAFE_NEW(pTrackTime,int,nTrackCount,1);
			nChunkUsed = min(nChunkSize / 4,nTrackCount);

			fread(pTrackTime,nChunkUsed,4,file);
			fseek(file,nChunkSize - (nChunkUsed * 4),SEEK_CUR);

			for(; nChunkUsed < nTrackCount; nChunkUsed++)
				pTrackTime[nChunkUsed] = -1;	//negative signals to use default time

			break;

		case CHUNKTYPE_FADE:
			if(!bInfoFound)						return -1;
			if(pTrackFade)						return -1;

			SAFE_NEW(pTrackFade,int,nTrackCount,1);
			nChunkUsed = min(nChunkSize / 4,nTrackCount);

			fread(pTrackFade,nChunkUsed,4,file);
			fseek(file,nChunkSize - (nChunkUsed * 4),SEEK_CUR);

			for(; nChunkUsed < nTrackCount; nChunkUsed++)
				pTrackFade[nChunkUsed] = -1;	//negative signals to use default time

			break;

		case CHUNKTYPE_BANK:
			if(bBankFound)						return -1;

			bBankFound = 1;
			nChunkUsed = min(8,nChunkSize);

			fread(nBankswitch,nChunkUsed,1,file);
			fseek(file,nChunkSize - nChunkUsed,SEEK_CUR);
			break;

		case CHUNKTYPE_PLST:
			if(pPlaylist)						return -1;

			nPlaylistSize = nChunkSize;
			if(nPlaylistSize < 1)				break;  //no playlist?

			SAFE_NEW(pPlaylist,BYTE,nPlaylistSize,1);
			fread(pPlaylist,nChunkSize,1,file);
			break;

		case CHUNKTYPE_AUTH:		{
			if(szGameTitle)						return -1;

			char*		buffer;
			char*		ptr;
			SAFE_NEW(buffer,char,nChunkSize + 4,1);

			fread(buffer,nChunkSize,1,file);
			ptr = buffer;

			char**		ar[4] = {&szGameTitle,&szArtist,&szCopyright,&szRipper};
			int			i;
			for(i = 0; i < 4; i++)
			{
				nChunkUsed = lstrlen(ptr) + 1;
				SAFE_NEW(ar[i],char,nChunkUsed,0);
				if(!*ar[i]) { SAFE_DELETE(buffer); return 0; }
				memcpy(*ar[i],ptr,nChunkUsed);
				ptr += nChunkUsed;
			}
			SAFE_DELETE(buffer);
									}break;

		case CHUNKTYPE_TLBL:		{
			if(!bInfoFound)						return -1;
			if(szTrackLabels)					return -1;

			SAFE_NEW(szTrackLabels,char*,nTrackCount,1);

			char*		buffer;
			char*		ptr;
			SAFE_NEW(buffer,char,nChunkSize + nTrackCount,1);

			fread(buffer,nChunkSize,1,file);
			ptr = buffer;

			int			i;
			for(i = 0; i < nTrackCount; i++)
			{
				nChunkUsed = lstrlen(ptr) + 1;
				SAFE_NEW(szTrackLabels[i],char,nChunkUsed,0);
				if(!szTrackLabels[i]) { SAFE_DELETE(buffer); return 0; }
				memcpy(szTrackLabels[i],ptr,nChunkUsed);
				ptr += nChunkUsed;
			}
			SAFE_DELETE(buffer);
									}break;

		default:		//unknown chunk
			nChunkType &= 0x000000FF;  //check the first byte
			if((nChunkType >= 'A') && (nChunkType <= 'Z'))	//chunk is vital... don't continue
				return -1;
			//otherwise, just skip it
			fseek(file,nChunkSize,SEEK_CUR);

			break;
		}		//end switch
	}			//end while

	//if we exited the while loop without a 'return', we must have hit an NEND chunk
	//  if this is the case, the file was layed out as it was expected.
	//  now.. make sure we found both an info chunk, AND a data chunk... since these are
	//  minimum requirements for a valid NSFE file

	if(!bInfoFound)			return -1;
	if(!nDataPos)			return -1;

	//if both those chunks existed, this file is valid.  Load the data if it's needed

	if(needdata)
	{
		fseek(file,nDataPos,SEEK_SET);
		SAFE_NEW(pDataBuffer,BYTE,nDataBufferSize,1);
		fread(pDataBuffer,nDataBufferSize,1,file);
	}
	else
		nDataBufferSize = 0;

	//return success!
	return 0;
}
