/* 
 * Copyright (C) 2004      Disch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_FME07.h
//
//

struct FME07Wave
{
	/////////////////////////////
	// Frequency Control
	union TWIN		nFreqTimer;
	int			nFreqCount;

	/////////////////////////////
	// Channel Disabling
	BYTE		bChannelEnabled;

	/////////////////////////////
	// Volume
	BYTE		nVolume;

	/////////////////////////////
	// Duty Cycle
	BYTE		nDutyCount;

	///////////////////////////////////
	// Output and Downsampling
	short		nOutputTable_L[0x10];
	short		nOutputTable_R[0x10];
	int			nMixL;
	int			nMixR;

	///////////////////////////////////
	// Inversion
	BYTE		bInvert;
	BYTE		bDoInvert;
	WORD		nInvertFreqCutoff;
};

extern struct FME07Wave mWave_FME07[3];
#define this mWave_FME07[chan]
	
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void Wave_FME07_DoTicks(int chan, int ticks,BYTE mix)
	{
		register int mn;

		if(!this.bChannelEnabled)		return;
		if(!this.nFreqTimer.W)			return;

		while(ticks)
		{
			mn = min(this.nFreqCount,ticks);
			ticks -= mn;

			this.nFreqCount -= mn;

			if(mix && (this.nDutyCount < 16))
			{
				this.nMixL += this.nOutputTable_L[this.nVolume] * mn;
				this.nMixR += this.nOutputTable_R[this.nVolume] * (this.bDoInvert ? -mn : mn);
			}

			if(this.nFreqCount > 0) continue;
			this.nFreqCount = this.nFreqTimer.W;

			if(++this.nDutyCount >= 32)
			{
				this.bDoInvert = this.bInvert && (this.nFreqTimer.W < this.nInvertFreqCutoff);
				this.nDutyCount = 0;
			}
		}
	}

	inline void Wave_FME07_Mix_Mono(int chan, int * mix,int downsample)
	{
		*mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_FME07_Mix_Stereo(int chan, int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);

		this.nMixL = this.nMixR = 0;
	}

#undef this
