/* 
 * Copyright (C) 2004      Disch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_MMC5.h
//
//    These are similar to the native NES square waves, only they lack a sweep unit and aren't
//  (as far as I know), interdependant on each other's output.
//
//    The voice is similar to the $4011 register, but without the DMC's sample playback capabilities.
//  so it's rather useless.  I haven't been able to find any game to test it with (since I'm not aware of
//  any who use it... nor do I see how it could be used in an NSF because of lack of IRQ support).  But it's
//  included anyway.  Theoretically it should work... but like I said, can't test it.

struct MMC5SquareWave
{

	///////////////////////////////////
	// Programmable Timer
	union TWIN		nFreqTimer;
	int			nFreqCount;

	///////////////////////////////////
	// Length Counter
	BYTE		nLengthCount;
	BYTE		bLengthEnabled;
	BYTE		bChannelEnabled;

	///////////////////////////////////
	// Volume / Decay
	BYTE		nVolume;
	BYTE		nDecayVolume;
	BYTE		bDecayEnable;
	BYTE		bDecayLoop;
	BYTE		nDecayTimer;
	BYTE		nDecayCount;

	///////////////////////////////////
	// Duty Cycle
	BYTE		nDutyCount;
	BYTE		nDutyCycle;

	///////////////////////////////////
	// Output and Downsampling
	BYTE		bChannelMix;
	short		nOutputTable_L[0x10];
	short		nOutputTable_R[0x10];
	int			nMixL;
	int			nMixR;

	///////////////////////////////////
	// Inverting
	BYTE		bDoInvert;
	BYTE		bInvert;
	WORD		nInvertFreqCutoff;
	
};


extern struct MMC5SquareWave mWave_MMC5Square[2];
#define this mWave_MMC5Square[chan]

	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void Wave_MMC5Square_ClockMajor(int chan)		//decay
	{
		if(this.nDecayCount)
			this.nDecayCount--;
		else
		{
			this.nDecayCount = this.nDecayTimer;
			if(this.nDecayVolume)
				this.nDecayVolume--;
			else
			{
				if(this.bDecayLoop)
					this.nDecayVolume = 0x0F;
			}

			if(this.bDecayEnable)
				this.nVolume = this.nDecayVolume;
		}
	}

	inline void Wave_MMC5Square_ClockMinor(int chan)		//length
	{
		if(this.bLengthEnabled && this.nLengthCount)
			this.nLengthCount--;
	}

	inline void Wave_MMC5Square_DoTicks(int chan, int ticks,BYTE mix)
	{
		register int mn;

		if(this.nFreqTimer.W < 8) return;

		while(ticks)
		{
			mn = min(this.nFreqCount,ticks);
			ticks -= mn;

			this.nFreqCount -= mn;

			if(mix && (this.nDutyCount < this.nDutyCycle) && this.nLengthCount)
			{
				this.nMixL += this.nOutputTable_L[this.nVolume] * mn;
				this.nMixR += this.nOutputTable_R[this.nVolume] * (this.bDoInvert ? -mn : mn);
			}

			if(!this.nFreqCount)
			{
				this.nFreqCount = this.nFreqTimer.W + 1;
				this.nDutyCount = (this.nDutyCount + 1) & 0x0F;
				if(!this.nDutyCount)
				{
					this.bDoInvert = this.bInvert;
					if(this.nInvertFreqCutoff < this.nFreqTimer.W)
						this.bDoInvert = 0;
				}
			}
		}
	}

	inline void Wave_MMC5Square_Mix_Mono(int chan, int * mix,int downsample)
	{
		*mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_MMC5Square_Mix_Stereo(int chan, int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);

		this.nMixL = this.nMixR = 0;
	}

#undef this

struct MMC5VoiceWave
{
	///////////////////////////////////
	// Everything
	BYTE		nOutput;
	short		nOutputTable_L[0x80];
	short		nOutputTable_R[0x80];
	int			nMixL;
	int			nMixR;
	BYTE		bInvert;
	
};

extern struct MMC5VoiceWave	mWave_MMC5Voice;
#define this mWave_MMC5Voice

	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////


	inline void Wave_MMC5Voice_DoTicks(int ticks)
	{
		this.nMixL += this.nOutputTable_L[this.nOutput] * ticks;
		this.nMixR += this.nOutputTable_R[this.nOutput] * (this.bInvert ? -ticks : ticks);
	}

	inline void Wave_MMC5Voice_Mix_Mono(int * mix,int downsample)
	{
		*mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_MMC5Voice_Mix_Stereo(int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);

		this.nMixL = this.nMixR = 0;
	}

#undef this
