/* 
 * Copyright (C) 2004      Disch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_N106.h
//
//    I love this chip!  Although it's not as easy as the others to emulate... but it sure sounds nice.
//


struct N106Wave
{
	//////////////////////////////
	//  All Channel Stuff
	BYTE		nActiveChannels;
	BYTE		bAutoIncrement;
	BYTE		nCurrentAddress;
	BYTE		nRAM[0x100];				//internal memory for registers/wave data
	float		fFrequencyLookupTable[8];	//lookup table for frequency conversions


	//////////////////////////////
	//  Individual channel stuff
	//////////////////////////////
	//  Wavelength / Frequency
	union QUAD		nFreqReg[8];
	float		fFreqTimer[8];
	float		fFreqCount[8];

	//////////////////////////////
	//  Wave data length / remaining
	BYTE		nWaveSize[8];
	BYTE		nWaveRemaining[8];

	//////////////////////////////
	//  Wave data position
	BYTE		nWavePosStart[8];
	BYTE		nWavePos[8];
	BYTE		nOutput[8];

	//////////////////////////////
	//  Volume
	BYTE		nVolume[8];

	//////////////////////////////
	//  Pop Reducer
	BYTE		nPreVolume[8];
	BYTE		nPopCheck[8];

	//////////////////////////////
	// Mixing
	short		nOutputTable_L[8][0x10][0x10];
	short		nOutputTable_R[8][0x10][0x10];
	int			nMixL[8];
	int			nMixR[8];

	//////////////////////////////
	// Inverting
	BYTE		bInvert[8];
	BYTE		bDoInvert[8];
	UINT		nInvertFreqCutoff[8][8];
	BYTE		nWaveSizeWritten[8];
	BYTE		nInvCheck[8];
};

extern struct N106Wave		mWave_N106;
#define this mWave_N106
	
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void Wave_N106_DoTicks(int ticks,BYTE* mix)
	{
		register int mn;
		register int i;
		int usetick;
		BYTE temp;

		for(i = (7 - this.nActiveChannels); i < 8; i++)
		{
			if(!this.nFreqReg[i].D)
			{
				// written frequency of zero will cause divide by zero error
				// makes me wonder if the formula was supposed to be Reg+1
				this.nVolume[i] = this.nPreVolume[i];
				this.bDoInvert[i] = this.bInvert[i];
				continue;
			}

			usetick = ticks;
			while(usetick)
			{
				mn = (int)ceil(this.fFreqCount[i]);
				mn = min(mn,usetick);

				usetick -= mn;
				if(mix[i])
				{
					this.nMixL[i] += this.nOutputTable_L[i][this.nVolume[i]][this.nOutput[i]] * mn;
					this.nMixR[i] += this.nOutputTable_R[i][this.nVolume[i]][this.nOutput[i]] * (this.bDoInvert[i] ? -mn : mn);
				}

				if(this.fFreqTimer[i] < 0)
					this.fFreqTimer[i] = (this.fFrequencyLookupTable[this.nActiveChannels] / this.nFreqReg[i].D);
				if(this.fFreqCount[i] > this.fFreqTimer[i])
					this.fFreqCount[i] = this.fFreqTimer[i];

				this.fFreqCount[i] -= mn;
				if(this.fFreqCount[i] <= 0)
				{
					this.fFreqCount[i] += this.fFreqTimer[i];
					if(this.nWaveRemaining[i])
					{
						this.nWaveRemaining[i]--;
						this.nWavePos[i]++;
					}
					if(!this.nWaveRemaining[i])
					{
						this.nWaveRemaining[i] = this.nWaveSize[i];
						this.nWavePos[i] = this.nWavePosStart[i];
						if(this.nVolume[i] != this.nPreVolume[i])
						{
							if(++this.nPopCheck[i] >= 2)
							{
								this.nPopCheck[i] = 0;
								this.nVolume[i] = this.nPreVolume[i];
							}
						}

						temp = this.bInvert[i] && (this.nInvertFreqCutoff[this.nActiveChannels][this.nWaveSizeWritten[i]] < this.nFreqReg[i].D);
						if(temp != this.bDoInvert[i])
						{
							if(++this.nInvCheck[i] >= 2)
							{
								this.nInvCheck[i] = 0;
								this.bDoInvert[i] = temp;
							}
						}
					}

					this.nOutput[i] = this.nRAM[this.nWavePos[i]];
					if(!this.nOutput[i])
					{
						this.nPopCheck[i] = 0;
						this.nVolume[i] = this.nPreVolume[i];
						this.bDoInvert[i] = this.bInvert[i] && (this.nInvertFreqCutoff[this.nActiveChannels][this.nWaveSizeWritten[i]] < this.nFreqReg[i].D);
					}
				}
			}
		}
	}

	inline void Wave_N106_Mix_Mono(int * mix,int downsample)
	{
		register int i;
		for(i = 0; i < 8; i++)
		{
			*mix += (this.nMixL[i] / downsample);
			this.nMixL[i] = 0;
		}
	}
	
	inline void Wave_N106_Mix_Stereo(int * mixL,int * mixR,int downsample)
	{
		register int i;
		for(i = 0; i < 8; i++)
		{
			*mixL += (this.nMixL[i] / downsample);
			*mixR += (this.nMixR[i] / downsample);
			this.nMixL[i] = this.nMixR[i] = 0;
		}
	}
#undef this
